"""Regular expressions for validating and parsing URLs"""

import re

# Much of this regex is quite similar to the django URL validator
# <https://github.com/django/django/blob/master/django/core/validators.py>
scheme = r"(?P<scheme>https?)://"
user_pass = r"(?P<user_pass>(?:[^\s:@/]+(?::[^\s:@/]*)?(?=@))?)@?"
characters = "a-z" + "\u00a1-\uffff" + "0-9"
ipv4 = r"(?:25[0-5]|2[0-4]\d|[0-1]?\d?\d)(?:\.(?:25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}"
ipv6 = r"\[[0-9a-f:.]+\]"
hostname = rf"[{characters}](?:[{characters}-]{{0,61}}[{characters}])?"
domain = rf"(?:\.(?!-)[{characters}-]{{1,63}}(?<!-))*"
tld = r"\.(?!-)(?:[a-z" + "\u00a1-\uffff" + r"-]{2,63}|xn--[a-z0-9]{1,59})(?<!-)\.?"
host = f"(?P<host>{hostname}{domain}{tld}|localhost)"
port = r":?(?P<port>(?:(?<=:)\d{2,5})?)"
resource = r"(?P<resource>(?:[/?#][^\s]*)?)"
url = scheme + user_pass + host + port + resource + r"\Z"

url_regex = re.compile(url, re.IGNORECASE)


def parse(url):
    """Parse the parts of a URL

    Parameters
    ----------
    url : str

    Returns
    -------
    match : dict
        If the URL is valid, this returns a dictionary with keys "scheme",
        "user_pass", "host", "port", and "resource" describing the various parts of
        the URL.  Note that if any of these parts are missing, the values of those
        keys will be empty strings.  If the URL is not valid, the dictionary will
        be empty.

    See Also
    --------
    validate : just return True or False depending on whether the URL is valid

    """
    match = url_regex.match(url)
    return getattr(match, "groupdict", {})


def validate(url):
    """Validate a URL

    Parameters
    ----------
    url : str

    Returns
    -------
    valid : bool

    See Also
    --------
    parse : parse the parts of a URL into a dictionary

    """
    return bool(url_regex.match(url))



"""Base class for all waveform-like objects"""

import abc
import numpy as np
import quaternionic


class WaveformMixin(abc.ABC):
    # Note: This is currently a pretty trivial class, but as the code develops
    # this will be increasingly useful.

    @property
    @abc.abstractmethod
    def t(self):  # Handy alias for backwards compatibility
        return self.time

    @property
    @abc.abstractmethod
    def data(self):  # Handy alias for backwards compatibility
        return self.ndarray

    @property
    @abc.abstractmethod
    def data_type(self):
        return self._metadata.get("data_type", "unknown")

    @property
    @abc.abstractmethod
    def spin_weight(self):
        return self._metadata.get("spin_weight", None)

    @property
    @abc.abstractmethod
    def frame(self):
        return self._metadata.get("frame", np.atleast_2d(quaternionic.one))

    @property
    @abc.abstractmethod
    def frame_type(self):
        if "frame" in self._metadata:
            return self._metadata.get("frame_type", "unknown")
        else:
            return self._metadata.get("frame_type", "inertial")

    # metadata
    # spin_weight
    # boost_weight
    # type
    #     name
    #     radius_scaling
    #     mass_scaling
    #     conformal_scaling
    #     frame ∈ {'inertial', 'corotating', ...}
    # copy()
    # evaluate(units, total_mass, theta, phi, orientation, time, ...)  # Return WaveformSignal



"""Functions to convert horizon quantities to LVC format"""

import numpy as np
import h5py


def prepare_horizon_quantity(sxs_horizon_quantity, start_time, peak_time):
    """Return times and values of an SXS-format horizon quantity.

    Horizon quantities are things like AhA.dir/ArealMass.dat. This function
    first truncates the horizon data, including only data after the reference
    time. Then, it shifts the time by the same amount as the waveforms (i.e.,
    by the peak time). Then, return the truncated/shifted times and truncated
    values.

    """
    # First, figure out the correct time series
    times_raw_AH = sxs_horizon_quantity[:, 0]
    start_AH = np.argmin(np.abs(times_raw_AH - start_time))
    times_AH = times_raw_AH[start_AH:] - peak_time

    # Loop over remaining components, truncating each one to match times_AH
    quantity_AH_list = []
    for i in range(1, len(sxs_horizon_quantity[0])):
        quantity_AH_list.append(sxs_horizon_quantity[start_AH:, i])
    quantity_AH = np.array(quantity_AH_list)
    return times_AH, quantity_AH


def spline_horizon_quantity(sxs_horizon_quantity, start_time, peak_time, truncation_tol=None):
    """Return spline of a horizon quantity.

    Prepares `sxs_horizon_quantity` by passing it to `prepare_horizon_quantity`
    and then returns a spline of the result.

    """
    from . import Dataset
    tol = 1e-6
    if truncation_tol is True:  # Test for actual identity, not just equality
        truncation_tol = 5e-2 * tol
    elif truncation_tol is False:
        truncation_tol = None
    times_AH, quantity_AH = prepare_horizon_quantity(sxs_horizon_quantity, start_time, peak_time)
    spline_AH_list = [
        Dataset.from_data(times_AH, quantity_AH[i], tol=tol, truncation_tol=truncation_tol)
        for i in range(0, len(sxs_horizon_quantity[0]) - 1)
    ]
    return spline_AH_list


def insert_spline(sxs_horizons, spline_dictionary, spline_keys,
                  horizon_key, quantity_key, start_time, peak_time, log=print, truncation_tol=None):
    """Insert spline of horizon quantity into dictionary of horizon splines.

    Note: spline_keys is a vector of key names (should be length 1 for scalars,
    length 3 for vectors), where each key is the name of a group that will be
    written in the LVC file, such as mass1-vs-time or spin1x-vs-time;
    horizon_key is AhA, AhB, or AhC; and quantity_key is, e.g.,
    ChristodoulouMass, the name of the quantity in Horizons.h5 to be read and
    splined.

    """
    log("Computing " + str(spline_keys))
    ah = str(horizon_key) + ".dir"
    qty = str(quantity_key) + ".dat"
    quantity = sxs_horizons[ah][qty]
    spline = spline_horizon_quantity(quantity, start_time, peak_time, truncation_tol=truncation_tol)
    for i, spline_key in enumerate(spline_keys):
        spline_dictionary[spline_key] = spline[i]


def derived_horizon_quantities_from_sxs(sxs_horizons, start_time, peak_time):
    """Compute nhat, omega_orbit, LNhat, and horizon times from Horizons.h5

    Specifically, returns a tuple containing nhat, a unit vector from the
    secondary black hole to the primary black hole; omega_orbit, the orbital
    frequency; LNhat, a unit vector in the direction of the orbital angular
    momentum; and the horizon times of the primary, secondary, and remnant,
    truncated and shifted.

    """
    t_A, x_A = prepare_horizon_quantity(
        sxs_horizons['AhA.dir']['CoordCenterInertial.dat'], start_time,
        peak_time)
    t_B, x_B = prepare_horizon_quantity(
        sxs_horizons['AhB.dir']['CoordCenterInertial.dat'], start_time,
        peak_time)
    # This is used only for t_C
    t_C, x_C = prepare_horizon_quantity(
        sxs_horizons['AhC.dir']['CoordCenterInertial.dat'], start_time,
        peak_time)

    # n_vec is a unit vector pointing from the secondary (black hole B) to
    # the primary (black hole A). We use np.linalg.norm() to get the Euclidean
    # magnitude of n_vec.
    x_A = x_A.T
    x_B = x_B.T
    n_vec = x_A-x_B
    n_vec_norm = np.linalg.norm(n_vec, axis=-1)
    n_hat = n_vec/n_vec_norm[:, None]

    # We compute dn_vec/dt to get a velocity vector, by computing differences
    # for dn_vec and dt.
    dn_vec = np.diff(n_vec, axis=0)
    dt = np.diff(t_A)
    dn_vec_dt = dn_vec/dt[:, None]

    # The orbital frequency is the magnitude of n_vec x dn_vec/dt
    # This is just from the Newtonian expression |r x v| = r^2 \omega.
    # Because taking a time derivative reduces the length of the array by 1,
    # drop the last time of n_vec so n_vec and dn_vec_dt have the same number
    # of points.
    r_cross_v = np.cross(n_vec[:-1], dn_vec_dt)
    r_cross_v_norm = np.linalg.norm(r_cross_v, axis=-1)
    omega_orbit = r_cross_v_norm / n_vec_norm[:-1]**2

    # Finally, LNhat is a unit vector in the direction of the orbital
    # angular momentum. That is, it is a unit vector in the direction of
    # r x p, which is the same direction as r x v.
    LN_hat = r_cross_v / r_cross_v_norm[:, None]

    # Horizons.h5 stores quantities as functions of time. Append time to the
    # derived quantities.
    n_hat_vs_time = np.c_[t_A, n_hat]
    omega_orbit_vs_time = np.c_[t_A[:-1], omega_orbit]
    LN_hat_vs_time = np.c_[t_A[:-1], LN_hat]
    return n_hat_vs_time, omega_orbit_vs_time, LN_hat_vs_time, t_A, t_B, t_C


def insert_derived_spline(spline_dictionary, spline_keys, derived_quantity, log=print, truncation_tol=None):
    """Inserts a spline into a dictionary of horizon splines

    Derived from Horizons.h5.  Note: spline_keys is a vector of key
    names (should be length 1 for scalars, length 3 for vectors), where
    each key is the name of a group that will be written in the LVC
    file, such as Omega-vs-time or LNhatx-vs-time; and derived_quantity
    is the quantity to be splined.  The derived_quantity should be
    computed using derived_horizon_quantities_from_sxs().

    """
    from . import Dataset
    # N.B. times already shifted, truncated to remove junk by
    # derived_horizon_quantities_from_sxs()
    log("Computing " + str(spline_keys))
    tol = 1e-6
    if truncation_tol is True:  # Test for actual identity, not just equality
        truncation_tol = 5e-2 * tol
    elif truncation_tol is False:
        truncation_tol = None
    times_AH = derived_quantity[:, 0]
    quantity_AH = derived_quantity[:, 1:]
    for i, spline_key in enumerate(spline_keys):
        spline_dictionary[spline_key] = Dataset.from_data(times_AH, quantity_AH[:, i], tol=tol, truncation_tol=truncation_tol)


def horizon_splines_from_sxs(horizons, start_time, peak_time, log=print, truncation_tol=None):
    """Prepare dictionary of horizon-quantity splines

    The LVC format expects such a dictionary.  This function creates
    one, starting with an SXS-format Horizons.h5. The start_time and
    peak_time are determined by waveforms.convert_modes.

    """
    horizon_splines = {}

    # Christodoulou mass
    insert_spline(horizons, horizon_splines, ['mass1-vs-time'],
                  'AhA', 'ChristodoulouMass', start_time, peak_time, log, truncation_tol=truncation_tol)
    insert_spline(horizons, horizon_splines, ['mass2-vs-time'],
                  'AhB', 'ChristodoulouMass', start_time, peak_time, log, truncation_tol=truncation_tol)
    insert_spline(horizons, horizon_splines, ['remnant-mass-vs-time'],
                  'AhC', 'ChristodoulouMass', start_time, peak_time, log, truncation_tol=truncation_tol)

    # Dimensionless spin
    insert_spline(horizons, horizon_splines,
                  ['spin1x-vs-time', 'spin1y-vs-time', 'spin1z-vs-time'],
                  'AhA', 'chiInertial', start_time, peak_time, log, truncation_tol=truncation_tol)
    insert_spline(horizons, horizon_splines,
                  ['spin2x-vs-time', 'spin2y-vs-time', 'spin2z-vs-time'],
                  'AhB', 'chiInertial', start_time, peak_time, log, truncation_tol=truncation_tol)
    insert_spline(horizons, horizon_splines,
                  ['remnant-spinx-vs-time', 'remnant-spiny-vs-time', 'remnant-spinz-vs-time'],
                  'AhC', 'chiInertial', start_time, peak_time, log, truncation_tol=truncation_tol)

    # Position
    insert_spline(
        horizons, horizon_splines,
        ['position1x-vs-time', 'position1y-vs-time', 'position1z-vs-time'],
        'AhA', 'CoordCenterInertial', start_time, peak_time, log, truncation_tol=truncation_tol)
    insert_spline(
        horizons, horizon_splines,
        ['position2x-vs-time', 'position2y-vs-time', 'position2z-vs-time'],
        'AhB', 'CoordCenterInertial', start_time, peak_time, log, truncation_tol=truncation_tol)
    insert_spline(
        horizons, horizon_splines,
        ['remnant-positionx-vs-time', 'remnant-positiony-vs-time', 'remnant-positionz-vs-time'],
        'AhC', 'CoordCenterInertial', start_time, peak_time, log, truncation_tol=truncation_tol)

    # Derived quantities: nhat, omega_orbit, LNhat
    n_hat, omega_orbit, LN_hat, t_A, t_B, t_C = derived_horizon_quantities_from_sxs(horizons, start_time, peak_time)
    insert_derived_spline(horizon_splines, ['nhatx-vs-time', 'nhaty-vs-time', 'nhatz-vs-time'], n_hat, log, truncation_tol=truncation_tol)
    insert_derived_spline(horizon_splines, ['Omega-vs-time'], omega_orbit, log, truncation_tol=truncation_tol)
    insert_derived_spline(horizon_splines, ['LNhatx-vs-time', 'LNhaty-vs-time', 'LNhatz-vs-time'], LN_hat, log, truncation_tol=truncation_tol)

    return horizon_splines, t_A, t_B, t_C


def write_horizon_splines_from_sxs(
        out_filename,
        horizon_splines,
        primary_horizon_times,
        secondary_horizon_times,
        remnant_horizon_times,
        log=print):
    """Takes a dictionary of horizon splines, prepared with
    horizon_splines_from_sxs, and writes each spline into an HDF5 file. Also
    outputs the horizon times for the individual and remnant black holes,
    truncated to remove junk radiation and shifted."""
    log("Writing horizon data to '{0}'".format(out_filename))
    with h5py.File(out_filename, 'a') as out_file:
        for key in horizon_splines.keys():
            out_group = out_file.create_group(key)
            horizon_splines[key].write(out_group)
        out_file.create_dataset('HorizonATimes', data=primary_horizon_times)
        out_file.create_dataset('HorizonBTimes', data=secondary_horizon_times)
        out_file.create_dataset('CommonHorizonTimes', data=remnant_horizon_times)



"""Simple regexes to understand SXS IDs"""

sxs_identifier_regex = r"(?P<sxs_identifier>SXS:(?P<simulation_type>BBH|BHNS|NSNS):(?P<sxs_number>[0-9]+))(?:v(?P<version>[0-9]+))?"
lev_regex = r"Lev(?P<lev>[-0-9]*)"


def sxs_id(s, default=""):
    """Return the SXS ID contained in the input string

    An SXS ID is anything that matches the following regular expression:

        SXS:(BBH|BHNS|NSNS):[0-9]*

    If no match is found, the empty string is returned.  If multiple matches are
    found, only the first is returned.

    The input parameter may be:
    1) A string believed to contain the SXS ID
    2) The path to a file (like common-metadata.txt) containing such a string
    3) An object with a 'title' attribute
    4) An object with a 'title' item

    """
    import os.path
    import re
    try:
        s = s["title"]
    except (TypeError, KeyError):
        pass
    if hasattr(s, "title") and not isinstance(s.title, type("a".title)):
        s = s.title
    try:
        if os.path.isfile(s):
            with open(s, "r") as f:
                s = [l.strip() for l in f.splitlines()]
            for line in s:
                sxs_id_line = sxs_id(line)
                if sxs_id_line:
                    return sxs_id_line
            return default
    except TypeError:
        pass
    m = re.search(sxs_identifier_regex, s)
    if m:
        return m["sxs_identifier"]
    else:
        return default


def simulation_title(sxs_id):
    """Create simulation title given just the SXS ID

    Currently supported types return output that looks like

        Binary black-hole simulation SXS:BBH:0001
        Black-hole neutron-star binary simulation SXS:BHNS:0001
        Binary neutron-star simulation SXS:NSNS:0001

    """
    import re
    sxs_system_re = re.compile(sxs_identifier_regex)
    m = sxs_system_re.search(sxs_id)
    if not m:
        raise ValueError(f"No SXS identifier found in '{sxs_id}'")
    sxs_identifier = m["sxs_identifier"]
    simulation_type = m["simulation_type"]
    if simulation_type == "BBH":
        title = f"Binary black-hole simulation {sxs_identifier}"
    elif simulation_type == "BHNS":
        title = f"Black-hole neutron-star binary simulation {sxs_identifier}"
    elif simulation_type == "NSNS":
        title = f"Binary neutron-star simulation {sxs_identifier}"
    else:
        raise ValueError(f"Did not recognize SXS system type '{simulation_type}'; should be BBH, BHNS, or NSNS.")
    return title


def lev_number(s):
    """Return the integer Lev number contained in the input string

    The Lev number is anything that matches the following regular expression:

        Lev(?P<lev>[-0-9]*)

    If no match is found, `None` is returned.  If multiple matches are found,
    only the first is returned.

    """
    import os.path
    import re
    m = re.search(lev_regex, s)
    if m:
        return int(m["lev"])
    else:
        return None



"""Various utilities used the sxs package"""

import functools
import numba

jit = functools.partial(numba.njit, cache=True)
vectorize = functools.partial(numba.vectorize, nopython=True, cache=True)
guvectorize = functools.partial(numba.guvectorize, nopython=True, cache=True)

# This is a generically reliable set of widths to feed to multishuffle when using 64-bit floats
default_shuffle_widths = (8, 8, 4, 4, 4, 2,) + (1,) * 34

from . import url, inspire, monotonicity, decimation, lvcnr
from .downloads import download_file
from .bitwise import xor, multishuffle
from .sxs_identifiers import sxs_identifier_regex, lev_regex, sxs_id, lev_number, simulation_title
from .sxs_directories import sxs_directory, read_config, write_config, sxs_path_to_system_path, cached_path
from .select import select_by_path_component
from .formats import file_format
from .pretty_print import fit_to_console
from .files import md5checksum, lock_file_manager, find_simulation_directories, find_files
from .dicts import KeyPassingDict



"""Class to contain a single LVC-NR-format dataset"""

class Dataset(object):
    """Represents a single dataset of a group in an LVC-format HDF5 file

    The lvcnrpy module requires all fields in format 3 to have subfields 'X', 'Y',
    'deg', 'tol', and 'errors'.  The first four of these make sense.  The last is
    not so relevant for LVC purposes; it is inherited from romspline, and is
    actually the L1 convergence measure of romspline.  In particular, it is nearly
    -- but distinctly not -- the same size as 'X' and 'Y'.  This is naturally
    useful for investigating the algorithm itself, but is irrelevant to using its
    results.  Moreover, since this class uses the "peak-greed" variant of the
    algorithm, that dataset no longer means the same thing.  But since it is
    required, we include here an `errors` member, containing a single element,
    being the largest error (scaled, if relevant) in the final result.

    """
    def __init__(self, *args, **kwargs):
        if args or kwargs:
            raise ValueError("This is an empty constructor; use `from_data` or `read` to add data.")

    @classmethod
    def from_data(cls, x, y, tol, rel=False, error_scaling=None, truncation_tol=None):
        """Construct reduced-order dataset from (x, y) data"""
        import numpy as np
        from ... import TimeSeries
        from ..decimation.peak_greed import minimal_grid
        lvc_dataset = cls()
        lvc_dataset.tol = tol
        y_reference = y
        if truncation_tol is not None:
            y = TimeSeries(y.copy(), x)
            y.truncate(truncation_tol)
            y = y.ndarray
        if error_scaling is None:
            error_scaling = 1.0
        indices = minimal_grid(x, y, tol=tol, error_scale=error_scaling, y_reference=y_reference)
        lvc_dataset.deg = 3
        lvc_dataset.X = x[indices].copy()
        lvc_dataset.Y = y[indices].copy()
        if error_scaling is None:
            lvc_dataset.errors = np.array([np.max(np.abs(y_reference - lvc_dataset.spline(x)))])
        else:
            lvc_dataset.errors = np.array([np.max(np.abs(error_scaling * (y_reference - lvc_dataset.spline(x))))])
        # lvc_dataset.compression_ratio = x.size/lvc_dataset.X.size
        # print("Size ratio:", x.size/lvc_dataset.X.size)
        return lvc_dataset

    def write(self, output_group):
        import h5py
        if not isinstance(output_group, h5py.Group):
            raise Exception("Parameter `output_group` must be an h5py.Group (or File) object.")
        output_group.create_dataset('deg', data=self.deg, dtype='int')
        output_group.create_dataset('tol', data=self.tol, dtype='double')
        output_group.create_dataset('errors', data=self.errors, dtype='double')
        output_group.create_dataset(
            'X', data=self.X, dtype='double', compression='gzip', shuffle=True, chunks=(self.X.size,)
        )
        output_group.create_dataset(
            'Y', data=self.Y, dtype='double', compression='gzip', shuffle=True, chunks=(self.Y.size,)
        )

    @classmethod
    def read(cls, input_group):
        import h5py
        if not isinstance(input_group, h5py.Group):
            raise Exception("Parameter `input_group` must be an h5py.Group (or File) object.")
        lvc_dataset = Dataset()
        lvc_dataset.deg = input_group['deg'][()]
        lvc_dataset.tol = input_group['tol'][()]
        lvc_dataset.errors = input_group['errors'][:]
        lvc_dataset.X = input_group['X'][:]
        lvc_dataset.Y = input_group['Y'][:]
        return lvc_dataset

    def spline(self, xprime):
        from scipy.interpolate import InterpolatedUnivariateSpline as spline
        return spline(self.X, self.Y, k=self.deg)(xprime)



class Login(object):

    def __init__(self, sandbox=False, access_token=None, access_token_path=None,
                 total_retry_count=50, backoff_factor=0.1, backoff_max=20.0, session=None):
        """Initialize a Login object for interacting with zenodo

        This object encapsulates the credentials needed to interact with the Zenodo API, and exposes
        a Session object that can be used to make requests which automatically include the
        credentials.  It can be used for generic requests, but note that other objects in this
        module make certain tasks easier -- such as creating or modifying a "deposit", which is
        Zenodo's name for a new upload.  The Deposit object should be created from this object.

        These actions require a Zenodo API access token.  These are obtained from the website --
        either
            https://zenodo.org/account/settings/applications/tokens/new/
        or
            https://sandbox.zenodo.org/account/settings/applications/tokens/new/
        Note that zenodo.org and sandbox.zenodo.org use separate login systems and separate access
        tokens.  The access token may either be passed as a string to this function (though that
        means it will probably be found in some script file somewhere, which is probably not a good
        idea for security), or can be read from a file.  By default the file from which the token is
        read is '~/.credentials/zenodo/zenodo_access_token' or the same name with '_sandbox' appended.
        Thus, it is probably easiest to simply place your access tokens in those files, so that no
        arguments need to be passed to this function.  As a basic security measure, please ensure
        that those files are not readable by anyone but the user.

        Parameters
        ----------
        sandbox: bool [default: False]
            If True, use the zenodo sandbox site, which is intended solely for testing purposes.
            This site is cleaned out regularly, so you cannot expect any entry to be here for very
            long.

        access_token: string or None [default: None]
            If present, this is used as the Zenodo API access token.

        access_token_path: string or None [default: None]
            If `zenodo_access_token` is not given, this file is read and the first line is used as the
            access token.  If this argument is None, it defaults to either
            '~/.credentials/zenodo/zenodo_access_token' for the regular website or
            '~/.credentials/zenodo/access_token_sandbox' for the sandbox website.

        total_retry_count: int [default: 50]
            Total number of times to retry requests that fail for retry-able reasons.

        backoff_factor: float [default: 0.1]
            A delay factor to apply between requests after the second try (most errors are resolved
            immediately by a second try without a delay).  After a certain number of total retries,
            the request Session will sleep for:

                {backoff factor} * (2 ^ ({number of total retries} - 1))

            seconds before trying again. For example, if the `backoff_factor` is 0.1, then the
            session will sleep for [0.0s, 0.2s, 0.4s, ...] between retries.  It will never be longer
            than `backoff_max`.

        backoff_max: float [default: 20.0]
            Longest time (in seconds) to wait between retries.

        session: requests.Session or None [default: None]
            This is the object that handles all of the requests made to the API.  If `None`, a
            Session is created for you, and sensible default headers (including the access token)
            are created.  If you need to adjust some of the Session parameters like proxies or SSL
            verification, you can simply create your own and pass it in here.  Note that any `auth`
            property on the passed object will be replaced by one that adds this to the header of
            each request to the chosen Zenodo domain:
                {"Authorization": "Bearer <YourAccessTokenHere>"}

        """
        import os
        import requests
        from requests.adapters import HTTPAdapter
        from urllib3.util.retry import Retry
        from . import url_sandbox, url_standard

        self.sandbox = sandbox
        if self.sandbox:
            self.base_url = url_sandbox
        else:
            self.base_url = url_standard

        # The Session object will handle all requests we make.
        self.session = session or requests.Session()

        # If the input session object succeeds at a `get`, we can skip a lot of the following
        r = self.session.get("{0}api/deposit/depositions".format(self.base_url))
        if r.status_code != 200:  # That's okay, we mostly expected this

            # Set the Zenodo API access token
            if access_token is not None:
                self.access_token = access_token
            else:
                if access_token_path is None:
                    if self.sandbox:
                        access_token_path = '~/.credentials/zenodo/access_token_sandbox'
                    else:
                        access_token_path = '~/.credentials/zenodo/access_token'
                path = os.path.expanduser(access_token_path)
                try:
                    with open(path, 'r') as f:
                        self.access_token = f.readline().strip()
                except IOError:
                    print('Unable to find the Zenodo access token needed to make a deposit.')
                    print('Failed to open file "{0}" for reading.'.format(path))
                    raise
                if not self.access_token:
                    print('The file "{0}" did not contain any text on the first line.'.format(path))
                    print('This is should be a Zenodo access token, which is need to make a Deposit.')
                    raise ValueError('Deposit requires a Zenodo access token')

            # Ensure that this session sends the Authorization header with every request to the base_url
            class ZenodoAuth(requests.auth.AuthBase):
                def __init__(self, base_url, zenodo_access_token):
                    self.base_url = base_url
                    self.access_token = zenodo_access_token
                    super(ZenodoAuth, self).__init__()
                def __call__(self, zenodo_r):
                    if zenodo_r.url.startswith(self.base_url):
                        zenodo_r.headers.update({"Authorization": "Bearer {0}".format(self.access_token)})
                    return zenodo_r
            self.session.auth = ZenodoAuth(self.base_url, self.access_token)

        # Note that some requests require different choices for 'Accept' and 'Content-Type'; these
        # are altered in the corresponding methods below.
        default_headers= {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        self.session.headers.update(default_headers)

        ## Retry automatically on certain types of errors
        Retry.BACKOFF_MAX = backoff_max  # Must be set on the class, not the instance
        retry = Retry(
            total=total_retry_count,
            backoff_factor=backoff_factor,
            status_forcelist=[500, 502, 503, 504,],
        )
        adapter = HTTPAdapter(max_retries=retry)
        self.session.mount(self.base_url, adapter)

        # Test to see if we can use the given access token
        url = "{0}api/deposit/depositions".format(self.base_url)
        r = self.session.get(url)
        if r.status_code != 200:
            if r.status_code == 401:
                print('The given Zenodo access token was not accepted by {0}.  Please ensure that it is still valid.'.format(self.base_url))
                print('Also note that zenodo.org and sandbox.zenodo.org use separate logins and separate access tokens.')
            else:
                print('An unknown error occurred when trying to access {0}.'.format(self.base_url))
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error

    def download(self, url, path):
        """Download large file efficiently

        Parameters
        ----------
        url: string
            The URL to download from.  Redirects are followed.
        path: string
            Relative or absolute path to the file in which the download will be stored.  If this is
            an existing directory or ends in a path separator, the "path" component of the URL will
            be used as the file name, and the full directory path will be created.

        """
        from shutil import copyfileobj
        from os import makedirs
        from os.path import split, exists, join, isdir
        from functools import partial
        from urllib.parse import urlparse
        url_path = urlparse(url).path
        if isdir(path):
            path = join(path, url_path[1:])
            directory, filename = split(path)
            if not exists(directory):
                makedirs(directory)
            local_filename = join(directory, filename)
        else:
            directory, filename = split(path)
            if not exists(directory):
                makedirs(directory)
            if not filename:
                filename = url_path
            local_filename = join(directory, filename)
        r = self.session.get(url, stream=True, allow_redirects=True)
        if r.status_code != 200:
            print('An error occurred when trying to access <{0}>.'.format(url))
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error
        r.raw.read = partial(r.raw.read, decode_content=True)
        # r.raw.decode_content = True
        with open(local_filename, 'wb') as f:
            copyfileobj(r.raw, f)
        return local_filename

    @property
    def new_deposit(self):
        """Create a new Deposit object using this login"""
        return self.deposit()

    def deposit(self, deposition_id=None, ignore_deletion=False):
        """Retrieve a deposit created with this login"""
        from .deposit import Deposit
        return Deposit(self, deposition_id, ignore_deletion)

    def search(self, q=None, status=None, sort=None, page=1, size=1000, all_versions=False, max_pages=10):
        """Return list of dictionaries describing each deposit created with this login

        It is possible to filter the results use the optional parameters.  Note that the web interface
        can sometimes be used to find search parameters by looking in the `search-hidden-params`
        parameter of the `invenio-search` tag.

        Example queries
        ---------------
        'title:"SXS:BBH:0003"'  # Finds titles with given string; use quotes for robustness
        'communities:sxs'  # Records in the 'sxs' Zenodo community
        'provisional_communities:sxs'  # Records awaiting approval by the community curator
        'owners: 38418'  # Find records by id number of owner

        Optional parameters
        -------------------
        q: string [optional]
            Search query, using Elasticsearch query string syntax.  See
            https://help.zenodo.org/guides/search/ for details.
        status: string, either 'draft' or 'published' [optional]
            Filter result based on deposit status.
        sort: string [optional]
            Sort order ('bestmatch' or 'mostrecent').  Prefix with minus to change from ascending to
            descending (e.g., '-mostrecent').
        page: int [optional, defaults to 1]
            Page number for pagination
        size: int [optional, defaults to 1000]
            Number of results to return per page.  Note that Zenodo (as of this writing) seems to
            place a hard limit of 9999 responses.  Anything more will result in an error.  Use
            multiple pages to get more results.
        all_versions: bool [optional, defaults to False]
            If True return all records, including older versions of published records.
        max_pages: int [optional, defaults to 10]
            If the query returns a number of records equal to `size`, it is evidently incomplete.
            This function will attempt to retrieve successive pages until the number of records is
            less than `size`.  If the query is still incomplete after this many pages, just return
            what we've got.

        """
        params={}
        if q is not None:
            params['q'] = q
        if status is not None and status in ['draft', 'published']:
            params['status'] = status
        if sort is not None and sort in ['bestmatch', 'mostrecent', '-bestmatch', '-mostrecent']:
            params['sort'] = sort
        params['page'] = page
        params['size'] = size
        if all_versions:
            params['all_versions'] = ''

        url = "{0}api/deposit/depositions".format(self.base_url)
        r = self.session.get(url, params=params)
        if r.status_code != 200:
            print('An unknown error occurred when trying to access {0}.'.format(url))
            print('The search parameters were "{0}"'.format(params))
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error

        json = r.json()
        if len(json) == size:
            page += 1
            if page > max_pages:
                print('Search is not yet complete after {0} pages; returning with what we have.'.format(max_pages))
                return json  # Note: This will percolate back up the recursion to include other results
            return json + self.search(q=q, sort=sort, page=page, size=size, all_versions=all_versions, max_pages=max_pages)

        return json

    def delete_untitled_empty_deposits(self):
        deleted_deposits = 0
        deposits = self.search()
        for d in deposits:
            try:
                if d['title'] == '':
                    d = self.deposit(d['id'], ignore_deletion=True)
                    if not d.files:
                        d.delete_deposit(confirmed=True)
                        deleted_deposits += 1
            except:
                pass
        print('Deleted {0} deposits'.format(deleted_deposits))

    def discard_all_drafts(self):
        discarded_drafts = 0
        deposits = self.search()
        for d in deposits:
            try:
                if d['state'] == 'inprogress':
                    d = self.deposit(d['id'], ignore_deletion=True)
                    d.discard()
                    discarded_drafts += 1
            except:
                pass
        print('Discarded {0} drafts'.format(discarded_drafts))

    def awaiting_approval(self, community_id):
        """List all records awaiting approval for the given community"""
        url = '{0}/api/records/?q=provisional_communities:{1}'.format(self.base_url, community_id)
        r = self.session.get(url)
        if r.status_code != 200:
            print('Unable to find any records for community {0}.'.format(community_id))
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error
        return r.json()

    def community_curate_accept(self, community_id, record_id):
        """Accept a record into the community"""
        url = "{0}/communities/{1}/curaterecord/".format(self.base_url, community_id)
        data = {"recid": int(record_id), "action": "accept"}
        r = self.session.post(url, json=data)
        if r.status_code != 200:
            print('Unable to accept record id {0} into community {1}; status code={2}.'.format(record_id, community_id, r.status_code))
            try:
                r_json = r.json()
                print('Response JSON:')
                print(r_json)
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error
        return r.json()

    def total_deposit_size(self, deposition_id=None, human_readable=True):
        import math

        def convert_size(size_bytes):
            if size_bytes == 0:
                return "0B"
            size_name = ("B  ", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB")
            i = int(math.floor(math.log(size_bytes, 1024)))
            p = math.pow(1024, i)
            s = round(size_bytes / p, 3)
            return "{0:>8.3f} {1}".format(s, size_name[i])

        if deposition_id is None:
            depositions = self.search()
        else:
            depositions = [self.deposit(deposition_id, ignore_deletion=True).representation]
        total_size = 0
        for deposition in depositions:
            id = deposition['id']
            d = self.deposit(id, ignore_deletion=True)
            d_total_size = sum([f['filesize'] for f in d.files])
            print('{1} in "{2}" (Zenodo ID {0})'.format(id, convert_size(d_total_size), d.title))
            total_size += d_total_size
        print('{0} in {1} deposits'.format(convert_size(total_size), len(depositions)))
        if human_readable:
            return convert_size(total_size)  # Note: the return type will be str
        else:
            return total_size  # Note: the return type will be int



"""Full documentation for the contents of the catalog.json file"""


# NOTE: This string is placed into the top of the catalog JSON file as a JSON string.  JSON strings
# are enclosed in double quotes, so it would quickly get ugly if we used double quotes within this
# description, even though python makes that easy.
catalog_file_description = """
    This JSON file has the following format.  Comments are, of course, not present (since JSON does not support
    comments).  Single quotes here are, of course, double quotes in the rest of the file (since JSON encloses
    strings in double quotes).  Anything inside <angle brackets> is just standing in for the relevant value.  An
    ellipsis ... indicates that the preceding structure can be repeated.  Also note that the metadata entries for
    simulations may not be present if the record on zenodo is closed-access; see catalog_private_metadata.json if
    you have access to those simulations, which will contain the missing information.  That file should be read
    and written automatically by functions in this module, so that the catalog dict returned will contain all
    available information.

    {
        'catalog_file_description': '<this description>',
        'modified': '<YYYY-MM-DDThh:mm:ss.ssssss>',  # UTC time of last-modified record in this file
        'records': {  # Includes *all* records published on Zenodo in the 'sxs' community, not just simulations
            '<doi_url>': {  # This is precisely the value of the 'doi_url' key below
                # More details about this 'representation' object at http://developers.zenodo.org/#depositions
                'conceptdoi': '10.5281/zenodo.<conceptrecid>',  # Permanent DOI for all versions of this record
                'conceptrecid': '<conceptrecid>',  # ~7-digit integer (as string) collectively identifying all versions of this record
                'created': '<YYYY-MM-DDThh:mm:ss.ssssss>',  # UTC time of creation of this record on Zenodo
                'doi': '10.5281/zenodo.<id>',  # Permanent DOI for this record
                'doi_url': 'https://doi.org/10.5281/zenodo.<id>',  # URL for permanent DOI of this record
                'id': <id>,  # ~7-digit integer uniquely identifying this record
                'links': {
                     'badge': 'https://zenodo.org/badge/doi/10.5281/zenodo.<id>.svg',
                     'bucket': 'https://zenodo.org/api/files/<uuid>',  # Base URL for file uploads and downloads
                     'conceptbadge': 'https://zenodo.org/badge/doi/10.5281/zenodo.<conceptrecid>.svg',
                     'conceptdoi': 'https://doi.org/10.5281/zenodo.<conceptrecid>',  # Permanent link to webpage for most-recent version
                     'discard': 'https://zenodo.org/api/deposit/depositions/<id>/actions/discard',  # API action to discard a draft
                     'doi': 'https://doi.org/10.5281/zenodo.<id>',  # Permanent URL for this version
                     'edit': 'https://zenodo.org/api/deposit/depositions/<id>/actions/edit',  # API action to edit this record
                     'files': 'https://zenodo.org/api/deposit/depositions/<id>/files',  # Only present for author
                     'html': 'https://zenodo.org/deposit/<id>',  # Webpage for this version
                     'latest': 'https://zenodo.org/api/records/<id>',  # API endpoint for most-recent version
                     'latest_html': 'https://zenodo.org/record/<id>',  # Webpage for most-recent version
                     'publish': 'https://zenodo.org/api/deposit/depositions/<id>/actions/publish',  # Only present for author
                     'record': 'https://zenodo.org/api/records/<id>',  # Only present for author
                     'record_html': 'https://zenodo.org/record/<id>',  # Webpage for this particular version; only present for author
                     'self': 'https://zenodo.org/api/deposit/depositions/<id>'
                },
                'metadata': {  # Note that this is Zenodo metadata, and is different from the SXS metadata found below
                    'access_right': '<access>',  # Can be 'open', 'closed', 'embargoed', or 'restricted'
                    'communities': [
                        {'identifier': '<community_name>'},  # Names may include 'sxs' and 'zenodo'
                        ...
                    ],
                    'creators': [
                        {
                            'name': '<name>',  # Name of this creator in the format Family name, Given names
                            'affiliation': '<affiliation>',  # (Optional) Affiliation of this creator
                            'orcid': '<orcid>',  # (Optional) ORCID identifier of this creator
                            'gnd': '<gnd>'  # (Optional) GND identifier of this creator
                        },
                        ...
                    ],
                    'description': '<description>',  # Text description of this record
                    'doi': '10.5281/zenodo.<id>',  # Permanent DOI of this record
                    'keywords': [
                        '<keyword>',  # Optional; this array may be empty
                        ...
                    ],
                    'license': '<license_type>',  # Usually 'CC-BY-4.0' for SXS
                    'prereserve_doi': {'doi': '10.5281/zenodo.<id>', 'recid': <id>},
                    'publication_date': '<YYYY-MM-DD>',  # Possibly meaningless date (UTC)
                    'title': '<title>',
                    'upload_type': 'dataset'
                },
                'modified': '<YYYY-MM-DDThh:mm:ss.ssssss>',  # (UTC) Last modification of this record (possibly just Zenodo metadata modified)
                'owner': <user_id>,  # ~5-digit integer identifying the user who owns this record
                'record_id': <id>,  # Same as 'id'
                'state': '<state>',  # Can be 'done', 'inprogress', 'error', 'unsubmitted', possibly others
                'submitted': <submitted>,  # True or false (always true for published records)
                'title': '<title>'  # Same as ['metadata']['title'],
                'files': [  # May not be present if this simulation is closed-access; see catalog_private_metadata.json as noted above
                    # See https://data.black-holes.org/waveforms/documentation.html for
                    # detailed descriptions of the *contents* of the files in each record.
                    {
                        'checksum': '<checksum>',  # MD5 checksum of file on Zenodo
                        'filename': '<filename>',  # Name of file; may contain slashes denoting directories
                        'filesize': <filesize>,  # Number of bytes in the file
                        'id': '<fileid>',  # A standard UUID (hexadecimal with characters in the pattern 8-4-4-4-12)
                        'links': {
                            'download': 'https://zenodo.org/api/files/<bucket>/<filename>',  # The URL to use to download this file
                            'self': 'https://zenodo.org/api/deposit/depositions/<deposition_id>/files/<fileid>'  # Ignore this
                        }
                    },
                    ...  # Other file descriptions in the order in which they were uploaded (not necessarily a meaningful order)
                ]
            },
            ...
        },
        'simulations': {  # Physical data (masses, spins, etc.) for all available SXS simulations
            '<sxs_id>': {  # The SXS ID is a string like SXS:BHNS:0001 or SXS:BBH:1234
                'url': '<URL>',  # The URL of the Zenodo 'conceptdoi' link, which *resolves to* the most-recent version
                #
                # NOTE: All of the following may be absent if this simulation is closed-access, or simply does not have metadata.
                #
                # Variable content describing (mostly) physical parameters of the system.  It's basically a
                # python-compatible version of the information contained in 'metadata.txt' from the
                # highest-resolution run in the most-recent version of this simulation.  That file is meant to
                # be more-or-less as suggested in <https://arxiv.org/abs/0709.0093>.  The conversion to a
                # python-compatible format means that keys like 'simulation-name' have had hyphens replaced by
                # underscores so that they can be used as variable names in python and any other sane language
                # (with apologies to Lisp).  As far as possible, values that are just strings in that file
                # have been converted into the relevant types -- like numbers, integers, and arrays.  Note
                # that some keys like eccentricity are sometimes numbers and sometimes the string '<number'
                # (meaning that the eccentricity is less than the number), which is necessarily a string.
                #
                # Below are just the first few keys that *may* be present.  Note that closed-access
                # simulations will have empty dictionaries here.
                #
                'simulation_name': '<directory_name>',  # This may be distinctly uninformative
                'alternative_names': '<sxs_id>',  # This may be a list of strings
                'initial_data_type': '<type>',  # Something like 'BBH_CFMS'
                'object_types': '<type>',  # Currently 'BHBH', 'BHNS', or 'NSNS'
                'number_of_orbits': <number>,  # This is a float, rather than an integer
                'reference_mass_ratio': <q>,  # Usually greater than 1 (exceptions are due to junk radiation)
                'reference_chi_eff': <chi_eff>,  # Dimensionless effective spin quantity
                'reference_chi1_perp': <chi1_perp>,  # Magnitude of component of chi1 orthogonal to 'reference_orbital_frequency'
                'reference_chi2_perp': <chi2_perp>,  # Magnitude of component of chi2 orthogonal to 'reference_orbital_frequency'
                'reference_mass1': <m2>,
                'reference_mass2': <m1>,
                'reference_dimensionless_spin1': [
                    <chi1_x>,
                    <chi1_y>,
                    <chi1_z>
                ],
                'reference_dimensionless_spin2': [
                    <chi2_x>,
                    <chi2_y>,
                    <chi2_z>
                ],
                'reference_eccentricity': <eccentricity>,  # A float or possibly a string containing '<' and a float
                'reference_orbital_frequency': [
                    <omega_x>,
                    <omega_y>,
                    <omega_z>
                ],
                'reference_time': <time>,
                ...
            },
            ...
        }
    }
"""



"""Select items from a list"""


def select_by_path_component(path_pattern, possible_matches, recursion_index=0):
    """Select from a list of paths by progressively matching path components

    Parameters
    ----------
    path_pattern : str
        A pattern to search for in the sequence of possible matches.  If
        `recursion_index` is 0, this is first searched as an exact partial match,
        meaning paths that start with exactly this string are chosen.  If that does
        not produce any matches, the pattern is split into components, separated by
        "/".  This is then treated as a series of patterns to match corresponding
        path components in the `possible_matches`.  Each component pattern will be
        interpreted first as a literal match, then — if no match is found — as a
        regex, compiled by python's `re` module.
    possible_matches : Iterable[str]
        A sequence of paths to select matches from.
    recursion_index : int, optional
        This is presumably mostly for use from inside this function, and denotes
        the number of path components to try to match at once.  Defaults to 0.

    Returns
    -------
    matched_paths : Set[str]
        
    Notes
    -----
    Here, we consider a path as a sequence of components separated by "/".  Each
    path component is just a string without any "/" character.

    This function progressively matches paths, starting with the first path
    component, then searching the remaining possibilities for the second path
    component, and so on.  In each search, the pattern is first treated as a
    literal string to match to *the beginning of* a path.  If nothing is found, the
    pattern is compiled as a regex (by python's standard `re` module), and the
    search is run again.  In each case, if the current component was only a partial
    match (meaning there are characters in this path component *after* whatever was
    matched by the pattern), then the "largest" result is used — meaning literally
    that the matching path components are fed into the `max` function.  The idea is
    that if there are, for example, multiple versions or multiple resolutions of a
    simulation, then the largest is usually the "best", so we can default to that.

    An example may help to clarify.  Suppose we have this list of possible matches:

        possible_matches = [
            "abc/lm/xyz",
            "abc/ln/xyz",
            "abd/lm/xyz",
            "abd/ln/xyz",
        ]

    Then, we select from them with this:

        >>> select_by_path_component("ab/ln/xyz", possible_matches)
        {"abd/ln/xyz"}

    Compare this with the result from

        >>> select_by_path_component("ab./ln/xyz", possible_matches)
        {"abc/ln/xyz", "abd/ln/xyz"}

    In the first case, only "ab" matched to the first component in any of
    possibilities, so the function narrowed down the list to just the largest of
    these: those paths where the first element is "abd".  In the second case, the
    "." did not literally match anything, so "ab." was compiled to a regex, which
    matched both "abc" and "abd" — which leave no remaining characters, and so
    taking the `max` element in each case does nothing.  On the other hand, if we
    had used a regex that only matched part of the first component, the `max` stage
    would still have eliminated some possibilities:

        >>> select_by_path_component("a./ln/xyz", possible_matches)
        {"abd/ln/xyz"}

    Only the "ab" portion was matched by this regex, so we continue searching only
    the paths starting with `max("abc", "abd")` — which is "abd".

    This process is repeated successively for each component, and any type of match
    may be used in any component.  For example, we can select with

        >>> select_by_path_component("ab./l./x", possible_matches)
        {"abc/lm/xyz", "abc/ln/xyz", "abd/lm/xyz", "abd/ln/xyz"}

    Here, both the first and second components use regexes.  Or we could use
    partial matching in the second component:

        >>> select_by_path_component("ab./l/x", possible_matches)
        {"abc/ln/xyz", "abd/ln/xyz"}

    Note that both "abc" and "abd" match for the first component, and their matches
    are evaluated separately when selecting the second component.  In both cases
    "ln" is the max matching component.

    """
    import collections
    import re

    # Look for exact matches
    if recursion_index == 0:
        matches = {p for p in possible_matches if p.startswith(path_pattern)}
        if matches:
            return matches

    # Separate the pattern into components
    split_path_pattern = path_pattern.split("/")

    # If we've searched all components of the pattern, return all possible matches
    if recursion_index > len(split_path_pattern):
        return possible_matches

    # Split the pattern into the part we're currently working on, and the rest
    path_working = "/".join(split_path_pattern[:recursion_index+1])
    path_suffix = "/".join(split_path_pattern[recursion_index+1:])
    if path_suffix:
        path_suffix = "/" + path_suffix

    # First, look for exact matches to beginning of string
    matches = collections.defaultdict(list)
    for f in possible_matches:
        if f.startswith(path_working):
            next_slash = f.find("/", f.index(path_working) + len(path_working) - 1)
            if next_slash < 0:
                next_slash = len(f)
            matched_component = f[:next_slash]
            matches[path_working].append((matched_component, f))

    # If that didn't give us anything, compile to a regex
    if not matches:
        path_regex = re.compile(path_working)
        component_regex = re.compile(path_working + r"[^/]*")
        for f in possible_matches:
            m = path_regex.match(f)
            if m:
                match = m.group()
                matched_component = component_regex.match(f).group()
                matches[match].append((matched_component, f))

    # Go through, finding the `max` for any group of partial matches
    best_matches = {}
    for match in matches:
        best_component = max(set(m[0] for m in matches[match]))
        best_possible_matches = [m[1] for m in matches[match] if m[0] == best_component]
        best_matches[best_component] = best_possible_matches

    # Recurse, if necessary
    selected = {
        path
        for best_component in best_matches
        for path in select_by_path_component(
            best_component + path_suffix,  # Search for found components place rest of pattern
            possible_matches=best_matches[best_component],
            recursion_index=recursion_index+1
        )
    }

    return selected



"""Class and function to convert SXS data to LVC-NR format"""

class SimulationConverter(object):
    class Log(object):
        """Object to replace `log` function that used global `history`

        Instead of using a global `history` variable, just create an instance of this
        class, and pass it around to any function that called the old `log` function.
        Just like that function, this instance can be called with a string and will
        print the string while storing all the strings passed to it.

        Functions expecting an instance of this class can also use `print` as a default
        argument, which will work the same, but not store the value.

        """
        def __init__(self, quiet):
            self.history = ""
            self.quiet = quiet

        def __call__(self, string):
            if not self.quiet:
                print(string)
            self.history += string + "\n"

        def __str__(self):
            return str(self.history)

        def __repr__(self):
            return repr(self.history)

    def __init__(self, modes=8, tolerance=1e-06, quiet=False):
        """Create an object to be used for converting many waveforms to LVC format

        Parameters
        ----------
        modes : {int, '22only'}, optional
            Modes to be placed in the output file.  Passing '22only' results in the
            (2,2) and (2,-2) modes being output.  Otherwise, each (l,m) mode up to and
            including the given integer value will be output.  Note that for backwards
            compatibility, 'all' is also supported, and is equivalent to the default
            value of `8`.
        tolerance : float, optional
            Target tolerance used in `sxs.utilities.greedy_spline.minimal_indices`.
        quiet : bool, optional
            If False (the default), echo each line of the log as it is created;
            otherwise just store the final log in the output file.

        """
        import os
        import time
        import json
        import platform
        import numpy
        import scipy
        import h5py
        import sxs

        self.modes = modes
        self.tolerance = tolerance
        self.quiet = quiet

        self.code_versions = (
            f"python=={platform.python_version()}\n"
            f"numpy=={numpy.version.version}\n"
            f"scipy=={scipy.version.full_version}\n"
            f"h5py=={h5py.version.version}\n"
            f"# h5py_api=={h5py.version.api_version}\n"
            f"# h5py_hdf5=={h5py.version.hdf5_version}\n"
            f"sxs=={sxs.__version__}\n"
        )

        self.command = (
            f"sxs.utilities.lvcnr.convert_simulation(\n"
            f"    sxs_data_path={{sxs_data_path!r}},\n"
            f"    out_path={{out_path!r}},\n"
            f"    truncation_time={{truncation_time!r}},\n"
            f"    resolution={{resolution!r}},\n"
            f"    modes={modes!r},\n"
            f"    tolerance={tolerance!r},\n"
            f"    quiet={quiet!r}\n"
            f")"
        )

        # Make sense of the `modes` parameter
        if modes == 'all':
            self.modes = [[l, m] for l in range(2, 9) for m in range(-l, l+1)]
        elif modes == '22only':
            self.modes = [[2, 2], [2, -2]]
        else:
            l_max = int(modes)
            self.modes = [[l, m] for l in range(2, l_max+1) for m in range(-l, l+1)]
        self.ell_max = max(lm[0] for lm in self.modes)

        # Load catalog metadata
        catalog = sxs.load("catalog")
        self.sxs_catalog = {
            'simulations': catalog.simulations,
            'records': catalog.records,
        }

        self.sxs_catalog_resolutions = sxs.zenodo.catalog.resolutions_for_simulations(self.sxs_catalog)


    def convert(self, sxs_data_path, out_path, truncation_time=None, resolution=None, truncation_tol=None):
        """Convert a simulation from the SXS BBH catalog into the LVC format.

        This function outputs a file in LVC format named SXS_BBH_####_Res#.h5 in
        out_path.

        Parameters
        ----------
        sxs_data_path : string
            Path to directory containing rhOverM_Asymptotic_GeometricUnits_CoM.h5,
            Horizons.h5, and metadata.json files.
        out_path : string
            Path where LVC-format file is to be output
        truncation_time : {None, float}
            If specified, truncate time series at this time instead of at the reference
            time
        resolution : {None, int}
            Integer giving the resolution (Lev) of the data to convert.  If this is not
            given, the resolution is determined automatically from sxs_data_path.
        truncation_tol : {None, bool, callable, float, array_like}, optional
            If None (the default) or False, nothing happens.  If True, the waveform
            data (amplitude and phase) are "truncated" so that bits with significance
            lower than `5e-2 * self.tolerance` are set to zero, for improved
            compression.  Any other input is passed to `sxs.TimeSeries.truncate`.  Note
            that this is not typically a very effective setting — perhaps providing
            another 10% compression; the output file sizes are dominated by fairly
            redundant time data unaffected by this parameter.

        """
        import os
        import time
        import json
        import h5py
        import sxs

        from .metadata import sxs_id_from_alt_names, write_metadata_from_sxs
        from .horizons import horizon_splines_from_sxs, write_horizon_splines_from_sxs
        from .waveforms import convert_modes

        log = self.Log(self.quiet)
        log(self.command.format(sxs_data_path=sxs_data_path, out_path=out_path,
                                truncation_time=truncation_time, resolution=resolution))
        log("Starting at "+time.strftime('%H:%M%p %Z on %b %d, %Y'))

        # Load metadata.json from this simulation
        with open(os.path.join(sxs_data_path, "metadata.json"), 'r') as f:
            metadata = json.load(f)

        # Determine the resolution of the input simulation, if needed
        if resolution is None:
            resolution = sxs.lev_number(sxs_data_path)
        if resolution is None:
            raise ValueError('No `resolution` value found in input arguments or data path.')

        sxs_id = sxs_id_from_alt_names(metadata['alternative_names'])
        log("Converting " + sxs_id)

        extrapolation_order = "Extrapolated_N2"
        log("Extrapolation order: " + extrapolation_order)

        out_name = out_path + "/" + sxs_id.replace(':', '_') + "_Res" + str(resolution) + ".h5"
        log("Output filename is '{0}'".format(out_name))

        start_time, peak_time, version_hist = convert_modes(
            sxs_data_path + "/rhOverM_Asymptotic_GeometricUnits_CoM.h5",
            metadata, out_name, self.modes, extrapolation_order, log,
            truncation_time, tolerance=self.tolerance/2.0, truncation_tol=truncation_tol
        )

        with h5py.File(sxs_data_path + "/Horizons.h5", 'r') as horizons:
            horizon_splines_to_write, t_A, t_B, t_C = horizon_splines_from_sxs(
                horizons, start_time, peak_time, log, truncation_tol=truncation_tol
            )
        write_horizon_splines_from_sxs(out_name, horizon_splines_to_write, t_A, t_B, t_C, log)

        write_metadata_from_sxs(out_name, resolution, metadata,
                                self.sxs_catalog, self.sxs_catalog_resolutions,
                                start_time, peak_time, self.ell_max, log)

        with h5py.File(out_name, 'a') as out_file:
            # Save information about versions of code used in this function
            out_file["auxiliary-info"].create_dataset('CodeVersions.txt', data=self.code_versions)

            # Copy VersionHist.ver into the new file, if available
            if version_hist is not None:
                log("Writing VersionHist.ver")
                out_file["auxiliary-info"].create_dataset('VersionHist.ver', data=version_hist)
            else:
                log("No VersionHist.ver found. Data being converted is version 0.")

            # Store the log output by this script as a dataset
            log("Finishing at "+time.strftime('%H:%M%p %Z on %b %d, %Y'))
            log("Writing log")
            out_file["auxiliary-info"].create_dataset('ConversionLog.txt', data=log.history)


def convert_simulation(sxs_data_path, out_path, truncation_time=None, resolution=None,
                       modes=8, tolerance=1e-06, quiet=False):
    """Convert a simulation from the SXS BBH catalog into the LVC format.

    This function outputs a file in LVC format named SXS_BBH_####_Res#.h5 in
    out_path.

    Note that this function is essentially a wrapper for
    `SimulationConverter.convert`.  If you have very many systems to convert, it is
    significantly faster to create the SimulationConverter object once, and then
    call the `convert` method for each system.

    Parameters
    ----------
    sxs_data_path : string
        Path to directory containing rhOverM_Asymptotic_GeometricUnits_CoM.h5, Horizons.h5,
        and metadata.json files.
    out_path : string
        Path where LVC format file is to be output
    truncation_time : {None, float}, optional
        If specified, truncate time series at this time instead of at the reference time
    resolution : {None, int}, optional
        Integer giving the resolution (Lev) of the data to convert.  If this is not given,
        the resolution is determined automatically from sxs_data_path.
    modes : {int, '22only'}, optional
        Modes to be placed in the output file.  Passing '22only' results in the (2,2)
        and (2,-2) modes being output.  Otherwise, each (l,m) mode up to and including
        the given l value will be output.  Note that for backwards compatibility, 'all'
        is also supported, and is equivalent to the default value of `8`.
    tolerance : float, optional
        Target tolerance used in `sxs.utilities.greedy_spline.minimal_indices`.
    quiet : bool, optional
        If False (the default), echo each line of the log as it is created; otherwise
        just store the final log in the output file.

    """
    lvc_converter = SimulationConverter(modes, tolerance, quiet)
    return lvc_converter.convert(sxs_data_path, out_path, truncation_time, resolution)
